# Copyright (c) 2023 NEC Corporation. All Rights Reserved.

import builtins
import collections
import importlib
import importlib.machinery
import inspect
import itertools
import os
import os.path

from . import _decorator
from ._decorator import nohook
from . import _util
from ._util import get_logger

import firefw.runtime

__all__ = ["nohook", "get_logger"]

SEP = "."
INIT = "__init__.py"
CACHE = "__pycache__"

# fmt: off
HookPair = collections.namedtuple(
    "HookPair", ["import_name", "as_name"]
)

HookEntry = collections.namedtuple(
    "HookEntry", ["import_name", "import_path", "as_name", "as_path"]
)
# fmt: on


# for ImportHook.ignore_names
def _get_default_ignore_names():
    suffixes = importlib.machinery.all_suffixes()
    result = list()

    top = os.path.dirname(__file__)
    gtor = os.walk(top, topdown=True, followlinks=False)
    for dirpath, dirnames, filenames in gtor:
        for filename in filenames:
            if any(filename.endswith(s) for s in suffixes):
                result.append(os.path.join(dirpath, filename))
        if CACHE in dirnames:
            dirnames.remove(CACHE)

    result.remove(_decorator.__file__)
    return result


class ImportHook:
    # genuine functions
    __import__ = builtins.__import__
    __importlib_import__ = importlib.__import__
    __import_module__ = importlib.import_module

    # filters for stack frames
    ignore_names = _get_default_ignore_names()
    ignore_words = ["frozen importlib"]

    def __init__(self, ignore_names=None, ignore_words=None):
        cls = self.__class__
        if inspect.currentframe() is None:  # pragma: no cover
            get_logger().warning(
                "WARNING: The function inspect.currentframe() is not "
                "available on the current Python environment. This may cause "
                "importhook to be slow."
            )

        self._hooks = dict()
        self._ignames = cls.ignore_names[:]
        self._ignames.extend(ignore_names or list())
        self._igwords = cls.ignore_words[:]
        self._igwords.extend(ignore_words or list())
        self._debug = cls._is_debug_enabled()

    @staticmethod
    def _is_debug_enabled():
        import logging

        return get_logger().isEnabledFor(logging.DEBUG)

    def add_hook(self, import_, as_=None):
        """Add import hook.

        `add_hook(import_='egg.spam', as_='spam')` is an analogy of
        `import egg.spam as spam`.

        By setting `import_='egg.spam', as_='spam'` and activating this
        hook, an import statement `import spam` behaves itself like
        `import egg.spam as spam`. If `as_` is `None`, the name of the
        most descendent module of `import_` is used for `as_`.

        """

        cls = self.__class__
        import_name = import_
        as_name = as_

        if as_name is None:
            left, sep, right = import_name.rpartition(SEP)
            if right != import_name:
                as_name = right
            else:
                raise ValueError(
                    f"'as_' is not given but 'import_' has "
                    f"no child module: {import_name!r}"
                )

        amodule = cls.__import_module__(as_name)
        imodule = cls.__import_module__(import_name)
        left, sep, right = import_name.partition(SEP)
        if left != import_name:
            imodule = cls.__import_module__(left)

        as_path = _util.removesuffix(amodule.__file__, INIT)
        as_path = os.path.abspath(as_path)
        import_path = _util.removesuffix(imodule.__file__, INIT)
        import_path = os.path.abspath(import_path)

        entry = HookEntry(import_name, import_path, as_name, as_path)
        self._hooks[entry.as_name] = entry

        get_logger().info(repr(entry))
        return self

    def activate(self, hook_import_module=False):
        def _targets():
            return [
                builtins.__import__,
                importlib.__import__,
                importlib.import_module,
            ]

        before_hook = tuple(map(repr, _targets()))

        builtins.__import__ = self
        importlib.__import__ = self
        if hook_import_module:
            importlib.import_module = self.import_module

        after_hook = tuple(map(repr, _targets()))

        logger = get_logger()
        for s, t in zip(before_hook, after_hook):
            if s != t:
                logger.info(f"{s} is hooked by {t}")

    @classmethod
    def deactivate(cls):
        builtins.__import__ = cls.__import__
        importlib.__import__ = cls.__importlib_import__
        importlib.import_module = cls.__import_module__
        get_logger().info("an import hook is deactivated")

    def is_active(self):
        x = builtins.__import__ is self
        y = importlib.__import__ is self

        if (x and not y) or (not x and y):
            logger = get_logger()
            logger.error(f"builtins.__import__:  {builtins.__import__!r}")
            logger.error(f"importlib.__import__: {importlib.__import__!r}")
            raise ValueError("invalid importhook state")

        return x and y

    def is_active_for_import_module(self):
        # NOTE: An id of a bound method may differ every time.
        # So even if `foo = obj.method` and `bar = obj.method`,
        # `foo is bar` may be False. Instead of it, an instance
        # `__self__` which binds the method is used.
        return getattr(importlib.import_module, "__self__", None) is self

    def _framefilter(self, filename):
        matches = any(s == filename for s in self._ignames)
        contains = any(s in filename for s in self._igwords)
        return not matches and not contains

    def _get_frames(self, skip=1):
        frames = firefw.runtime.get_frames(skip)
        filenames = filter(self._framefilter, map(inspect.getfile, frames))
        return map(os.path.abspath, filenames)

    def _to_be_hooked(self, name):
        assert name in self._hooks
        entry = self._hooks[name]

        return all(
            os.path.commonpath([descendant, ancestor]) != ancestor
            for descendant, ancestor in itertools.product(
                self._get_frames(2),
                [entry.as_path, entry.import_path, _decorator.__file__],
            )
        )

    def __call__(self, name, globals=None, locals=None, fromlist=(), level=0):
        """A wrapper for the built-in function `__import__`."""

        cls = self.__class__

        if name in self._hooks and self._debug:
            import pprint

            get_logger().debug(pprint.pformat(list(self._get_frames())))

        if name in self._hooks and self._to_be_hooked(name):
            entry = self._hooks[name]
            get_logger().info(f"hook: {name!r} -> {entry.import_name!r}")
            return cls.__import_module__(entry.import_name)

        return cls.__import__(name, globals, locals, fromlist, level)

    def import_module(self, name, package=None):
        """A wrapper for the function `importlib.import_module`."""

        cls = self.__class__

        if name in self._hooks and self._to_be_hooked(name):
            entry = self._hooks[name]
            get_logger().info(f"hook: {name!r} -> {entry.import_name!r}")
            return cls.__import_module__(entry.import_name, package)

        return cls.__import_module__(name, package)

    @classmethod
    def _genuine_builtin_import(cls):
        return cls.__import__

    @classmethod
    def _genuine_importlib_import(cls):
        return cls.__importlib_import__

    @classmethod
    def _genuine_import_module(cls):
        return cls.__import_module__


def activate_hook_pairs(pairs, hook_import_module=False):
    """Activate pairs of import hooks."""

    hook = ImportHook()
    for import_name, as_name in pairs:
        hook.add_hook(import_name, as_name)

    hook.activate(hook_import_module)
    return hook


def activate_hook(import_, as_=None, hook_import_module=False):
    """Activate an import hook.

    `activate_hook(import_='egg.spam', as_='spam')` is an analogy of
    `import egg.spam as spam`.

    After this hook is activated, an import statement `import spam`
    behaves itself like `import egg.spam as spam`. If `as_` is `None`,
    the name of the most descendent module of `import_` is used for
    `as_`. For example, `import_='egg.bacon.spam.sausage.without.spam'`
    makes `as_='spam'`.

    By default, the built-in function `__import__` and the function
    `importlib.__import__` are hooked. If `hook_import_module` is
    `True`, the function `importlib.import_module` is also hooked.

    """

    return activate_hook_pairs([HookPair(import_, as_)], hook_import_module)


def deactivate_hook():
    ImportHook.deactivate()


def _get_current_hook():
    hook = builtins.__import__
    if isinstance(hook, ImportHook):
        return hook
    else:
        return None
